#ifndef WALK_HH_
#define WALK_HH_


#include "ell/State.hh"
#include "ell/NeighborGenerator.hh"
#include "ell/StateAcceptor.hh"
#include "ell/WalkAbortionCriterion.hh"
#include "ell/StateCollector.hh"

namespace ell {
	
	
	
	/*!
	 * This function implements a random walk starting from State s
	 * doing at most "length" steps
	 * by executing ell::walk.
	 * 
	 * @author Daniel Maticzka
	 */
//	StateCollector* walkRandom(
//			const State* const s,
//			const unsigned int length,
//			StateCollector* scAccepted,
//			StateCollector* scChecked = NULL);

	
	
	//!	Walk is an abstract class which defines:
	//! - a static function "walk" which takes all neccesary params for a generic walk
	//!		This static function Walk::walk is then invoked by
	//!	- a virtual member function "walk" which only takes a state to start and a walk collector
	//!  This virtual function then usually(not mandatory) invokes the static function Walk::walk 
	//!  with fixed params depending on the type of walk(=the class type), e.g. RandomWalk has 
	//!  member "length" and therefore creates a special WalkAbortionCriterion..
	//!  In addition every child of the Walk class has a static member which essentially does the same as 
	//!  the virtual member function(usually invoking Walk::walk) but with the use of additional params 
	//!  instead of members
	//!
	//!  @author: Hannes Kochni, Daniel Maticzka, Martin Mann
	class Walk {
	public:
		
		//! virtual destructor
			virtual ~Walk() {};
			
		//! pure virtual function, s is start state 
		//!
		//! @param start the begin of the walk
		//! @param scWalk the StateCollector each step of the walk is added to
		//! @param scRejected the StateCollector each rejected possible step is
		//!                   added to
		//! @return the modified scWalk StateCollector
		virtual 
			StateCollector*
			walk(	const State* const start, 
					StateCollector* const scWalk, 
					StateCollector* scRejected = NULL
				) const = 0;
		
		//! This function implements a generic walk starting from State start.
		//! The properties of this search are defined by the 
		//! implementations of NeighborhoodGenerator, StateAcceptor
		//! WalkAbortionCriterion and StateCollector.
		//!
		//! @param start the begin of the walk
		//! @param scWalk the StateCollector each step of the walk is added to
		//! @param ng the NeighborGenerator used to get the neighbor list of the
		//!           last walk State to generate the next possible steps
		//! @param sa the StateAcceptor that decides if a step is accepted or
		//!           rejected
		//! @param wac the walk abortion criterion that decides if the end of
		//!            the walk has been reached or a further step is to do
		//! @param scRejected the StateCollector each rejected possible step is
		//!                   added to
		//! @return the modified scWalk StateCollector
		static 
			StateCollector* 
			walk(	const State* const start, 
					StateCollector* const scWalk,
					const NeighborGenerator* const ng, 
					const StateAcceptor* const sa, 
					const WalkAbortionCriterion* const wac, 
					StateCollector* scRejected = NULL
				);
		
	protected:
	};
	
	
	
	

	//!	WalkAdaptive defines:
	//!  - an enum for the kind of adaptive walk
	//!  - a static function "walkAdaptive" which takes all neccesary params for an adaptive walk
	//!	 - a virtual member function "walk" which does the same as "walkAdaptive" but substitutes
	//!  the NeighborGentype of "walkAdaptive" with a member
	//!
	//!  @author: Hannes Kochni, Daniel Maticzka, Martin Mann
	class WalkAdaptive : public Walk {
	public:
		
		//! defines the kinds of Adaptive Walks: with random neighbor or succesive neighborgeneration
			enum NeighborGenType {SUCCESSIVE, RANDOM};
			
		//! constructor for a WalkAdaptive object, takes the type of walk
		//!
		//! @param ngt the type of the neighbor generator to use
		//! @param isDegenerate whether or not a degenerate landscape is in use
		//!                and a 'lesser energy' check is not sufficient and 
		//!                the lexicographical order of the compressed 
		//!                representation is used as a tiebreaker (equal E).
		WalkAdaptive(	const NeighborGenType ngt,
						const bool isDegenerate = false
					);
		
		//! virtual destructor
		virtual ~WalkAdaptive() {};
		
		//! This static function implements an adaptive walk starting from State
		//! start by executing Walk::walk.
		//!
		//! @param start the begin of the walk
		//! @param scWalk the StateCollector each step of the walk is added to
		//! @param ngt the type of the neighbor generator to use
		//! @param scRejected the StateCollector each rejected possible step is
		//!                   added to
		//! @param isDegenerate whether or not a degenerate landscape is in use
		//!                and a 'lesser energy' check is not sufficient and 
		//!                the lexicographical order of the compressed 
		//!                representation is used as a tiebreaker (equal E).
		//! @return the modified scWalk StateCollector
		static 
			StateCollector* 
			walkAdaptive(
				const State* const start, 
				StateCollector* const scWalk,
				const NeighborGenType ngt,
				StateCollector* const scRejected = NULL,
				const bool isDegenerate = false
				);

		// implements abstract function of superclass
		virtual
			StateCollector* 
			walk(
				const State* const start, 
				StateCollector* const scWalk,
				StateCollector* const scRejected = NULL
				) const;
				
	protected:
		//! WalkAdaptive object stores which kind neighbor generation is used 
		const NeighborGenType t;

		//! whether or not a degenerate landscape is in use
		//! and a 'lesser energy' check is not sufficient and 
		//! the lexicographical order of the compressed 
		//! representation is used as a tiebreaker (equal E).
		const bool isDeg;
	};

	
	
	
	//!	WalkGradient defines:
	//!  - a static function "walkGradient" which takes all neccesary params for an gradient walk
	//!	 - a virtual member function "walk" which does what "walkGradient" does
	//!
	//!  @author: Hannes Kochni, Daniel Maticzka, Martin Mann
	class WalkGradient : public Walk{
	public:
			
		//! constructs a gradient walk wrapper object
		//! @param isDegenerate whether or not a degenerate landscape is in use
		//!                and a 'lesser energy' check is not sufficient and 
		//!                the lexicographical order of the compressed 
		//!                representation is used as a tiebreaker (equal E).
		WalkGradient( const bool isDegenerate = false );

		//! virtual destructor
		virtual ~WalkGradient() {};

		//! Static function to perform a gradient (steepest decent) walk.
		//!
		//! @param start the begin of the walk
		//! @param scWalk the StateCollector each step of the walk is added to
		//! @param scRejected the StateCollector each rejected possible step is
		//!                   added to
		//! @param isDegenerate whether or not a degenerate landscape is in use
		//!                and a 'lesser energy' check is not sufficient and 
		//!                the lexicographical order of the compressed 
		//!                representation is used as a tiebreaker (equal E).
		//! @return the modified scWalk StateCollector
		static 
			StateCollector* 
			walkGradient(
				const State* const start, 
				StateCollector* const scWalk,
				const bool isDegenerate = false
				);
		
		// implementation of virtual member function for a gradient walk of an WalkGradient object
		virtual 
			StateCollector* 
			walk(
				const State* const start, 
				StateCollector* const scWalk,
				StateCollector* const scRej
				) const;
		
	protected:
		
		//! whether or not a degenerate landscape is in use
		//! and a 'lesser energy' check is not sufficient and 
		//! the lexicographical order of the compressed 
		//! representation is used as a tiebreaker (equal E).
		const bool isDeg;
	};

	
	

	//!	WalkMC is a class for a Monte Carlo Walk, it defines:
	//!  - a static function "walkMC" which takes all neccesary params for an monte carlo walk
	//!	 - a virtual member function "walk" which does the same as "walkMC" but substitutes
	//!  parameters WalkAbortionCriterion and beta of "walkGradient" with members
	//!
	//!  @author: Hannes Kochni, Daniel Maticzka, Martin Mann
	class WalkMC : public Walk{
	public:
		
		//! constructor which takes an AbortionCriterion and a term beta(1/k*T) TODO (formula correct?) 
		//! which is used by the StateAcceptor(based on Metropolitan criterion)
		WalkMC(	const WalkAbortionCriterion* const wac, 
					const double beta );
		
		//! virtual destructor		
		virtual ~WalkMC() {delete wac;};
		
		//! Static function for a Monte Carlo walk using the Metropolis
		//! criterion parametrized by beta.
		//!
		//! @param start the begin of the walk
		//! @param scWalk the StateCollector each step of the walk is added to
		//! @param wac the walk abortion criterion that decides if the end of
		//!            the walk has been reached or a further step is to do
		//! @param beta beta parameter of metropolis criterion (beta=(1/k*T))		
		//! @param scRejected the StateCollector each rejected possible step is
		//!                   added to
		//! @return the modified scWalk StateCollector
		static 
			StateCollector* 
			walkMC(	const State* const start, 
					StateCollector* const scWalk, 
					const WalkAbortionCriterion* const wac, 
					const double beta,
					StateCollector* const scRejected = NULL
					);
		
		// implementation of virtual member function for a monte carlo walk of an WalkMC object
		virtual 
			StateCollector* 
			walk(
				const State* const start, 
				StateCollector* const scWalk,
				StateCollector* const scRejected = NULL
			) const;
				
	protected:
		
		//! WalkMC object stores the WalkAbortionCriterion used by the virtual walk function
		const WalkAbortionCriterion* const wac;	
			
		//! WalkMC object stores beta used by the virtual walk function, beta=(1/k*T) 		
		const double beta;
	};
	
	
	
	
	

	//!	WalkRandom is a class for a length restricted random walk, it defines:
	//!  - a static function "walkRandom" which takes all neccesary params for an monte carlo walk
	//!	 - a virtual member function "walk" which does the same as "walkRandom" but substitutes
	//!  the parameter length of "walkRandom" with a member (see constructor)
	//!
	//!  @author: Hannes Kochni, Daniel Maticzka, Martin Mann
	class WalkRandom : public Walk {
	public:
		
		//! constructor which takes the length of the random walk 	
		//! @param sa the StateAcceptor that decides if a step is accepted or
		//!           rejected
		//! @param length length of the random walk		
			WalkRandom(	
						const StateAcceptor* const sa, 
						const size_t length
					);
			
		//! virtual destructor		
			virtual ~WalkRandom() {};
			
		//! static function for a random walk
			//! @param start the begin of the walk
			//! @param scWalk the StateCollector each step of the walk is added to
			//! @param sa the StateAcceptor that decides if a step is accepted or
			//!           rejected
			//! @param length length of the random walk		
			//! @param scRejected the StateCollector each rejected possible step is
			//!                   added to
			//! @return the modified scWalk StateCollector
		static 
			StateCollector* 
			walkRandom( const State* const start, 
						StateCollector* const scWalk,
						const StateAcceptor* const sa, 
						const size_t length,
						StateCollector* const scRejected = NULL
						);
		
		// implementation of virtual member function for a random walk of a WalkRandom object
		virtual 
			StateCollector* 
			walk(	const State* const start, 
					StateCollector* const scWalk,
					StateCollector* const scRejected = NULL
					) const;
		
	protected:
		
		//! the StateAcceptor that decides if a step is accepted or rejected
		const StateAcceptor* const sa;
		
		//! length of the random walk, used by the virtual walk function
		const size_t length;
	};

	
	
} // end namespace ell

#endif /*WALK_HH_*/
