#ifndef RNAFREEENERGY_HH_
#define RNAFREEENERGY_HH_


#include <biu/RNAStructure.hh>

namespace ell
{

		/**
		 * This subclass of RNAStructure implements the free energy as
		 * the underlying energy function with the constant alphabet ACGU
		 * and the constant valid RNA base pairs AU, CG and GU (and their
		 * reverse).
		 * 
		 * @author Martin Mann, Sebastian Will, Andreas Richter
		 */
	class RNAFreeEnergy : public biu::RNAStructure
	{
	protected:
			/*! The common BasePair object of all RNAFreeEnergy objects.
			 *  The valid RNA base pairs are AU, CG und GU and their reverse.
			 */
		static const biu::AllowedBasePairs		COMMON_BPAIRS;

			/*! The common Alphabet object of all RNAFreeEnergy objects.
			 *  It contains the elements A, C, G and U.
			 */
		static const biu::Alphabet				COMMON_ALPHABET;
		
		  //! dummy value for energy to inform that recalculation is necessary
		static const double ENERGY_INF; 
		
		  //! free energy of the molecule calculated using the Vienna RNA package
		mutable double energy;

	public:
			/*! A single move (open or close one bond) on a RNA structure
		 	 *  at positions SingleMove.first and SingleMove.second.
		 	 */
		typedef std::pair<size_t, size_t> SingleMove;

		RNAFreeEnergy(	const std::string& rnaSeqStr,
						const std::string& rnaStructBracketDotStr);

		RNAFreeEnergy(	biu::Sequence* rnaSeq,
						const biu::Structure* const rnaStructBracketDot,
						const bool seqIsShared = false);

		RNAFreeEnergy(const RNAFreeEnergy& rnaFreeEnerg);
		virtual ~RNAFreeEnergy();
		
			/*! Applies inplace a single move at the positions
			 *  given by move.
			 *  In case of invalid moves, the behaviour
			 *  of the program is not specified.
			 */
		void applySingleMoveInPlace(const SingleMove& move);
		
			/*! Returns whether a move produces a valid RNA structure.
			 *  A move can only be valid if the first move position
			 *  move.first is less than the second move position
			 *  move.second!
			 */
		bool isValidSingleMove(const SingleMove& move) const;
		
			/*! Returns the next position for a closing bond
			 *  after pos which guarantes that the nested
			 *  property will be preserved.
			 * 
			 *  Returns biu::RNAStructure::INVALID_INDEX if there is
			 *  no next valid position available.
			 */
		size_t nextCompatibleSingleMovePos(const size_t pos) const;

		static const biu::Alphabet* getSequenceAlphabet() {
			return &COMMON_ALPHABET;
		}

	// abstract functions (BioMolecule)
		
			/*! Returns the free energy of this RNA structure
			 *  allowing AU, CG and GU base base pairs.
			 */
		virtual double	getEnergy() const;

		RNAFreeEnergy&	operator= (const RNAFreeEnergy& rnaStruct2);

		//! Sets the RNA structure in bracket notation.
		//! @param str the structure to set in bracket notation of correct 
		//!         length
	virtual void setStructure(const biu::Structure& str) {
		  // set structure
		RNAStructure::setStructure(str);
		  // inform that molecule has changed and e.g. energy has to be 
		  // recalculated
		moleculeChanged();
	}

	protected:
		
		  //! Informs the molecule that the sequence or structure has been 
		  //! changed to force recalculation of lazy properties (e.g. energy)
		virtual void moleculeChanged();
	};

} // namespace biu

#endif /*RNAFREEENERGY_HH_*/
